import pytest
from gusty import create_dag
from datetime import timedelta
from airflow.sensors.external_task import ExternalTaskSensor
from airflow.operators.empty import EmptyOperator

###############
## FIXTURES ##
###############


@pytest.fixture(scope="session")
def external_deps_dir():
    return "tests/dags/external_deps"


@pytest.fixture(scope="session")
def custom_sensor_class():
    class CustomExternalTaskSensor(ExternalTaskSensor):
        is_custom_sensor = True

    return CustomExternalTaskSensor


@pytest.fixture(scope="session")
def dag(external_deps_dir, custom_sensor_class):
    dag = create_dag(
        external_deps_dir,
        default_args={"email": "default@gusty.com", "retries": 5},
        task_group_defaults={"prefix_group_id": True},
        wait_for_defaults={"poke_interval": 12},
        wait_for_class=custom_sensor_class,
    )
    return dag


@pytest.fixture(scope="session")
def empty_dag(external_deps_dir):
    dag = create_dag(
        external_deps_dir,
        default_args={"email": "default@gusty.com", "retries": 5},
        task_group_defaults={"prefix_group_id": True},
        wait_for_defaults={"poke_interval": 12},
        wait_for_class=EmptyOperator,
    )
    return dag


###########
## TESTS ##
###########


def test_dag_level_ext_dep(dag):
    wait_for_task = dag.task_dict["wait_for_DAG_top_level_external"]
    assert wait_for_task.__dict__["execution_delta"] == timedelta(minutes=88)


def test_task_group_level_ext_dep(dag):
    wait_for_task = dag.task_dict["wait_for_task_group_external_some_task"]
    assert wait_for_task.__dict__["execution_delta"] == timedelta(minutes=89)


def test_task_level_ext_dep(dag):
    wait_for_task = dag.task_dict["wait_for_task_level_external_another_task_1"]
    assert wait_for_task.__dict__["execution_delta"] == timedelta(minutes=95)
    assert wait_for_task.__dict__["retries"] == 95


def test_custom_sensor(dag, custom_sensor_class):
    wait_for_task = dag.task_dict["wait_for_task_group_external_some_task"]
    assert isinstance(wait_for_task, custom_sensor_class)
    assert wait_for_task.is_custom_sensor


def test_empty_dag(empty_dag):
    wait_for_task = empty_dag.task_dict["wait_for_task_group_external_some_task"]
    assert isinstance(wait_for_task, EmptyOperator)



import os, sys, pkgutil, itertools, importlib
from airflow.version import version as _airflow_version
from inflection import underscore

############
## Params ##
############

airflow_version = int(_airflow_version.split(".")[0])

###########################
## Operator Import Logic ##
###########################


def get_operator_location(operator_string):
    """
    Get package name / determine if "local" keyword is used
    """
    # location will generally be 'airflow',
    # but if it's 'local', then we will look locally for the operator
    return operator_string.split(".")[0]


def get_operator_name(operator_string):
    """
    Get operator class
    """
    # the actual name of the operator
    return operator_string.split(".")[-1]


def get_operator_module(operator_string):
    """
    Get module name
    """
    # the module, for when the operator is not a local operator
    operator_path = ".".join(operator_string.split(".")[:-1])
    assert len(operator_path) != 0, (
        "Please specify a format like 'package.operator' to specify your operator. You passed in '%s'"
        % operator_string
    )
    return operator_path


# Add $AIRFLOW_HOME/operators directory to path for local.operator syntax to work

gusty_home = os.environ.get("GUSTY_HOME", "")
if gusty_home == "":
    CUSTOM_OPERATORS_DIR = os.path.join(os.environ.get("AIRFLOW_HOME", ""), "operators")
else:
    CUSTOM_OPERATORS_DIR = os.path.join(gusty_home, "operators")

sys.path.append(CUSTOM_OPERATORS_DIR)
module_paths = [("", [CUSTOM_OPERATORS_DIR])]
pairs = [
    [(_.name, m + _.name) for _ in pkgutil.iter_modules(path)]
    for m, path in module_paths
]
module_dict = dict(itertools.chain(*pairs))


def get_operator(operator_string, operator_cache):
    """
    Given an operator string, determine the location of that operator and return the operator object
    """

    operator = operator_cache.get(operator_string)
    if operator:
        return operator

    operator_name = get_operator_name(operator_string)
    operator_location = get_operator_location(operator_string)

    module_name = (
        module_dict[underscore(operator_name)]
        if operator_location == "local"
        else get_operator_module(operator_string)
    )

    operator = getattr(importlib.import_module(module_name), operator_name)
    operator_cache.update({operator_string: operator})
    return operator



import os


def get_dags_directory(dags_dir):
    current_directory_name = os.path.split(os.getcwd())[-1]

    # Check if current, parent, or child directory is the dags directory
    if current_directory_name == "dags":
        dags_dir = "."
    elif os.path.exists("dags"):
        dags_dir = "dags"
    elif os.path.exists("../dags"):
        dags_dir = ".."
    else:
        dags_dir = "."
    return dags_dir



from absql.files.loader import generate_loader as generate_absql_loader
from datetime import datetime, timedelta
from gusty.utils import days_ago

default_constructors = {
    "!days_ago": days_ago,
    "!timedelta": timedelta,
    "!datetime": datetime,
}


def generate_loader(custom_constructors=None):
    """Generates a SafeLoader with both standard Airflow and custom constructors"""
    dag_yaml_tags = default_constructors.copy()
    custom_constructors = custom_constructors or {}

    if isinstance(custom_constructors, list) and len(custom_constructors) > 0:
        custom_constructors = {
            ("!" + func.__name__): func for func in custom_constructors
        }

    if len(custom_constructors) > 0:
        dag_yaml_tags.update(custom_constructors)

    return generate_absql_loader(dag_yaml_tags)


def handle_user_constructors(user_constructors):
    if isinstance(user_constructors, list):
        user_constructors = {("!" + func.__name__): func for func in user_constructors}
    return user_constructors



import pytest
from gusty import create_dags

##############
## FIXTURES ##
##############


@pytest.fixture(scope="session")
def multiple_dags_dir():
    return "tests/dags/multiple_dags"


@pytest.fixture(scope="session")
def multi_serial(multiple_dags_dir):
    dag_dict = {}
    create_dags(
        multiple_dags_dir,
        dag_dict,
        schedule="0 11 * * *",
        default_args={"email": "default@gusty.com", "owner": "default"},
        wait_for_defaults={"timeout": 679},
    )
    return dag_dict


@pytest.fixture(scope="session")
def multi_concurrent(multiple_dags_dir):
    dag_dict = {}
    create_dags(
        multiple_dags_dir,
        dag_dict,
        concurrent=True,
        schedule="0 11 * * *",
        default_args={"email": "default@gusty.com", "owner": "default"},
        wait_for_defaults={"timeout": 679},
    )
    return dag_dict


###########
## Tests ##
###########


def test_wait_for_defaults(multi_serial, multi_concurrent):
    serial_a = (
        multi_serial["dag_a"].task_dict["wait_for_dag_b_task_2"].__dict__["timeout"]
    )
    concurrent_a = (
        multi_concurrent["dag_a"].task_dict["wait_for_dag_b_task_2"].__dict__["timeout"]
    )
    assert serial_a == 679
    assert concurrent_a == 679


def test_kwargs(multi_serial, multi_concurrent):
    serial_c = multi_serial["dag_c"].__dict__["schedule_interval"]
    concurrent_c = multi_concurrent["dag_c"].__dict__["schedule_interval"]
    assert serial_c == "0 11 * * *"
    assert concurrent_c == "0 11 * * *"


def test_kwargs_override(multi_serial, multi_concurrent):
    # a
    serial_a = multi_serial["dag_a"].__dict__["schedule_interval"]
    concurrent_a = multi_concurrent["dag_a"].__dict__["schedule_interval"]
    # b
    serial_b = multi_serial["dag_b"].__dict__["schedule_interval"]
    concurrent_b = multi_concurrent["dag_b"].__dict__["schedule_interval"]
    # a
    assert serial_a == "0 0 * * *"
    assert concurrent_a == "0 0 * * *"
    # b
    assert serial_b == "0 0 * * *"
    assert concurrent_b == "0 0 * * *"


def test_default_args(multi_serial, multi_concurrent):
    serial_c_owner = multi_serial["dag_c"].task_dict["task_1"].__dict__["owner"]
    serial_c_email = multi_serial["dag_c"].task_dict["task_1"].__dict__["email"]
    concurrent_c_owner = multi_concurrent["dag_c"].task_dict["task_1"].__dict__["owner"]
    concurrent_c_email = multi_concurrent["dag_c"].task_dict["task_1"].__dict__["email"]
    assert serial_c_owner == "default"
    assert serial_c_email == "default@gusty.com"
    assert concurrent_c_owner == "default"
    assert concurrent_c_email == "default@gusty.com"


def test_default_args_override(multi_serial, multi_concurrent):
    # a
    serial_a_owner = multi_serial["dag_a"].task_dict["task_1"].__dict__["owner"]
    serial_a_email = multi_serial["dag_a"].task_dict["task_1"].__dict__["email"]
    concurrent_a_owner = multi_concurrent["dag_a"].task_dict["task_1"].__dict__["owner"]
    concurrent_a_email = multi_concurrent["dag_a"].task_dict["task_1"].__dict__["email"]
    # b
    serial_b_owner = multi_serial["dag_b"].task_dict["task_1"].__dict__["owner"]
    serial_b_email = multi_serial["dag_b"].task_dict["task_1"].__dict__["email"]
    concurrent_b_owner = multi_concurrent["dag_b"].task_dict["task_1"].__dict__["owner"]
    concurrent_b_email = multi_concurrent["dag_b"].task_dict["task_1"].__dict__["email"]
    # a
    assert serial_a_owner == "a"
    assert serial_a_email == "a@gusty.com"
    assert concurrent_a_owner == "a"
    assert concurrent_a_email == "a@gusty.com"
    # b
    assert serial_b_owner == "b"
    assert serial_b_email == "b@gusty.com"
    assert concurrent_b_owner == "b"
    assert concurrent_b_email == "b@gusty.com"



import collections.abc
from pendulum import today


def days_ago(n):
    """Get a datetime object representing n days ago."""
    n = -1 * n
    return today("UTC").add(days=n)


def nested_update(d, u):
    d = d.copy()
    for k, v in u.items():
        if isinstance(v, collections.abc.Mapping):
            d[k] = nested_update(d.get(k, {}), v)
        else:
            d[k] = v
    return d



import os
from copy import copy
from functools import partial
from absql.utils import get_function_arg_names
from gusty.parsing.loaders import generate_loader
from gusty.parsing.parsers import parse_generic, parse_py, parse_ipynb, parse_sql
from gusty.utils import nested_update

default_parsers = {
    ".yml": parse_generic,
    ".yaml": parse_generic,
    ".Rmd": parse_generic,
    ".py": parse_py,
    ".ipynb": parse_ipynb,
    ".sql": parse_sql,
}


def parse(
    file_path,
    parse_dict=default_parsers,
    loader=None,
    runner=None,
    render_on_create=False,
):
    """
    Reading in yaml specs / frontmatter.
    """

    if loader is None:
        loader = generate_loader()

    path, extension = os.path.splitext(file_path)

    parser = parse_dict[extension]
    if "loader" in get_function_arg_names(parser):
        parser = partial(parser, loader=loader)
    if "runner" in get_function_arg_names(parser):
        parser = partial(parser, runner=runner)
    if "render_on_create" in get_function_arg_names(parser):
        parser = partial(parser, render_on_create=render_on_create)

    yaml_file = parser(file_path)

    assert "operator" in yaml_file, "No operator specified in yaml spec " + file_path

    # gusty always supplies a task_id and a file_path in a spec
    yaml_file["task_id"] = (
        os.path.splitext(os.path.basename(file_path))[0].lower().strip()
    )
    assert (
        yaml_file["task_id"] != "all"
    ), "Task name 'all' is not allowed. Please change your task name."

    yaml_file["file_path"] = file_path

    # gusty will also attach the absql_runner
    yaml_file["absql_runner"] = copy(runner)

    # Check dependencies
    if "dependencies" in yaml_file.keys():
        assert isinstance(
            yaml_file["dependencies"], list
        ), "dependencies needs to be a list of strings in {file_path}".format(
            file_path=file_path
        )
        assert all(
            [isinstance(dep, str) for dep in yaml_file["dependencies"]]
        ), "dependencies needs to be a list of strings in {file_path}".format(
            file_path=file_path
        )

    # Handle multi_task
    multi_task_spec = yaml_file.get("multi_task_spec") or {}

    python_callable = yaml_file.get("python_callable")
    python_callable_partials = yaml_file.get("python_callable_partials")
    if python_callable and python_callable_partials:
        for task_id, partial_kwargs in python_callable_partials.items():
            callable_update = {
                "python_callable": partial(python_callable, **partial_kwargs)
            }
            if task_id in multi_task_spec.keys():
                multi_task_spec[task_id].update(callable_update)
            else:
                multi_task_spec.update({task_id: callable_update})

    multi_specs = []
    if len(multi_task_spec) > 0:
        for task_id, spec in multi_task_spec.items():
            base_spec = yaml_file.copy()
            base_spec["task_id"] = task_id
            base_spec = nested_update(base_spec, spec)
            multi_specs.append(base_spec)

    if len(multi_specs) > 0:
        return multi_specs

    return yaml_file



import pytest
from gusty import create_dag
from gusty.errors import NonexistentDagDirError

##############
## FIXTURES ##
##############


@pytest.fixture(scope="session")
def nonexistent_dir():
    return "tests/dagz/nonexistent"


###########
## TESTS ##
###########


def test_missing_dag_error(nonexistent_dir):
    with pytest.raises(NonexistentDagDirError) as e:  # noqa
        create_dag(nonexistent_dir)


def test_missing_dag_message(nonexistent_dir):
    try:
        create_dag(nonexistent_dir)
    except NonexistentDagDirError as e:
        assert str(e) == "DAG directory tests/dagz/nonexistent does not exist"



import pytest
from gusty.utils.context import (
    generate_runner_context,
    generate_loader_constructors,
    generate_user_defined_macros,
)

# (Everything in order of priority)
# The loader (!) needs:
#   - Airflow user-defined macros
#   - any additional dag_constructors
#   - default constructors
# The runner ({{}}) needs:
#   - Airflow user-defined macros
#   - any additional dag_constructors
#   - Airflow var.value, var.json, and macros
#   - default constructors
# Airflow user-defined macros needs:
#   - Airflow user-defined macros
#   - any additional dag_constructors


@pytest.fixture()
def user_provided_macros():
    return {"udf": lambda: "udf", "var": lambda: "var"}


@pytest.fixture()
def dag_constructors():
    return {
        "!dc": lambda: "dc",
        "!udf": lambda: "nonono",
        "!datetime": lambda: "yesyesyes",
    }


@pytest.fixture()
def loader_constructors(user_provided_macros, dag_constructors):
    return generate_loader_constructors(user_provided_macros, dag_constructors)


@pytest.fixture()
def runner_context(loader_constructors):
    return generate_runner_context(loader_constructors)


@pytest.fixture()
def user_defined_macros(user_provided_macros, dag_constructors):
    return generate_user_defined_macros(user_provided_macros, dag_constructors)


def test_runner_context_macros_over_constructors(runner_context):
    # user_provided_macros are preferred over dag_constructors
    assert runner_context["udf"]() == "udf"


def test_runner_context_constructors_in_context(runner_context):
    # dag_constructors make it into runner context
    assert runner_context["dc"]() == "dc"


def test_runner_context_dag_con_over_default_con(runner_context):
    # dag_constructors are prefered over default_constructors
    assert runner_context["datetime"]() == "yesyesyes"


# Cannot include macros because Airflow uses deepcopy
# def test_runner_context_macros_in_context(runner_context):
#     # airflow_default_macros make it into runner context
#     assert runner_context["macros"].ds_add("2022-01-01", 1) == "2022-01-02"


def test_runner_context_udf_over_default_macros(runner_context):
    # user_defined_macros are peffered over airflow_default_macros
    assert runner_context["var"]() == "var"


def test_user_defined_macros_macros_over_constructors(user_defined_macros):
    # user_provided_macros are preferred over dag_constructors
    assert user_defined_macros["udf"]() == "udf"


def test_user_defined_macros_constructors_in_macros(user_defined_macros):
    # dag_constructors make it into user_defined_macros
    assert user_defined_macros["dc"]() == "dc"



dag_contents_map = {
    "hi.py": (
        "# ---\n"
        "# python_callable: say_hi\n"
        "# ---\n"
        "\n"
        "def say_hi():\n"
        '  greeting = "hi"\n'
        "  print(greeting)\n"
        "  return greeting\n"
    ),
    "hey.sql": (
        "---\n"
        "operator: airflow.providers.sqlite.operators.sqlite.SqliteOperator\n"
        "---\n"
        "\n"
        "SELECT 'hey'\n"
    ),
    "hello.yml": (
        "operator: airflow.operators.bash.BashOperator\n" "dependencies:\n",
        "  - hi\n",
        "bash_command: echo hello\n",
    ),
    "METADATA.yml": (
        'description: "Saying hello using different file types"\n',
        "doc_md: |-\n",
        "  This is a longform description,\n",
        "  which can be accessed from Airflow's\n",
        "  Graph view for your DAG. It looks\n",
        "  like a tiny poem.\n",
        'schedule: "0 0 * * *"\n',
        "catchup: False\n",
        "default_args:\n",
        "  owner: You\n",
        "  email: you@you.com\n",
        "  start_date: !days_ago 28\n",
        "  email_on_failure: True\n",
        "  email_on_retry: False\n",
        "  retries: 1\n",
        "  retry_delay: !timedelta \n",
        "    minutes: 5\n",
        "tags:\n",
        "  - docs\n",
        "  - demo\n",
        "  - hello\n",
    ),
}

create_dag_file = lambda dag_name: (
    f"create_{dag_name}.py",
    (
        "import os\n",
        "from gusty import create_dag\n",
        "\n",
        "# There are many different ways to find Airflow's DAGs directory.\n",
        f'# gusty_dag_dir returns something like: "/usr/local/airflow/dags/{dag_name}"\n',
        f'gusty_dag_dir = os.path.join(os.environ.get("AIRFLOW_HOME"), "dags/{dag_name}")\n'
        "\n",
        "gusty_dag = create_dag(gusty_dag_dir, latest_only=False)\n",
    ),
)

create_dags_file = lambda dag_folder_name: (
    f"create_{dag_folder_name}.py",
    (
        "import os\n",
        "from gusty import create_dags\n",
        "from gusty.utils import days_ago\n",
        "\n",
        f'# gusty_dags_dir returns something like: "/usr/local/airflow/dags/{dag_folder_name}"\n',
        f'gusty_dags_dir = os.path.join(os.environ.get("AIRFLOW_HOME"), "dags/{dag_folder_name}")\n',
        "\n",
        "create_dags(\n",
        "  gusty_dags_dir,\n",
        "  globals(),\n",
        '  schedule="0 0 * * *",\n',
        "  catchup=False,\n",
        "  default_args={\n",
        '    "owner": "you",\n',
        '    "email": "you@you.com",\n',
        '    "start_date": days_ago(1)\n',
        "  },\n",
        "  wait_for_defaults={\n",
        '    "mode": "reschedule"\n',
        "  },\n",
        '  extra_tags=["gusty_dags"],\n',
        "  latest_only=False\n" ")\n",
    ),
)



import click
import os

from gusty.cli.utils import get_dags_directory
from gusty.cli.sample_tasks import create_dag_file, dag_contents_map, create_dags_file


@click.group()
def cli():
    pass


@click.command()
@click.argument("func", type=click.Choice(["create-dag", "create-dags"]))
@click.option("--name", "-n", type=str)
@click.option("--dags-dir", "-d", type=click.Path())
def use(func, name, dags_dir):
    if not name:
        name = "hello_dag" if func == "create-dag" else "gusty_dags"

    # dags directory defaults to current directory if not specified
    # and is not the current, parent, or child directory.
    dags_dir = dags_dir if dags_dir else get_dags_directory(dags_dir)

    dag_path = os.path.join(dags_dir, name)
    if func == "create-dags":
        dag_path = os.path.join(dags_dir, name, "hello_dag")
        try:
            os.makedirs(dag_path)
        except FileExistsError:
            click.echo(f"DAG directory {name} already exists, exiting!", err=True)
            exit(1)
    else:
        try:
            os.mkdir(dag_path)
        except FileExistsError:
            click.echo(f"DAG {name} already exists, exiting!", err=True)
            exit(1)

    for filename, contents in dag_contents_map.items():
        fpath = os.path.join(dag_path, filename)
        with open(fpath, "x") as f:
            f.writelines(contents)

    create_dag_filename, create_dag_contents = (
        create_dag_file(name) if func == "create-dag" else create_dags_file(name)
    )

    create_file_fpath = os.path.join(dags_dir, create_dag_filename)
    with open(create_file_fpath, "x") as f:
        f.writelines(create_dag_contents)


cli.add_command(use)



import os
import asyncio
from gusty import create_dag


async def create_dag_async(dag_dir, **kwargs):
    return {
        "dag_id": os.path.basename(dag_dir),
        "dag": create_dag(dag_dir, **kwargs),
    }


async def create_dags_async(
    gusty_dags, caller_env, concurrent_timeout, max_concurrency, **kwargs
):
    async_results = asyncio.as_completed(
        [create_dag_async(gusty_dag, **kwargs) for gusty_dag in gusty_dags],
        timeout=concurrent_timeout,
    )
    sem = asyncio.Semaphore(max_concurrency)
    async with sem:
        for async_dag in async_results:
            try:
                dag_dict = await async_dag
                caller_env[dag_dict["dag_id"]] = dag_dict["dag"]
            except asyncio.exceptions.TimeoutError:
                raise asyncio.exceptions.TimeoutError(
                    f"create_dags took more than {concurrent_timeout} seconds. "
                    "To fix, adjust the concurrent_timeout in the call to create_dags, "
                    "or set a higher AIRFLOW__CORE__DAGBAG_IMPORT_TIMEOUT in your environment."
                )


def create_dags(
    dags_dir,
    caller_env,
    concurrent=False,
    concurrent_timeout=None,
    max_concurrency=8,
    current_dag_id=None,
    **kwargs,
):
    # assumes any subdirectories in the dags directory are
    # gusty DAGs (excludes subdirectories like __pycache__)
    gusty_dags = [
        os.path.join(dags_dir, name)
        for name in os.listdir(dags_dir)
        if os.path.isdir(os.path.join(dags_dir, name)) and not name.endswith("__")
    ]

    if concurrent:
        if concurrent_timeout is None:
            concurrent_timeout = (
                float(os.environ.get("AIRFLOW__CORE__DAGBAG_IMPORT_TIMEOUT", 30.0))
                * 0.9
            )
        asyncio.run(
            create_dags_async(
                gusty_dags, caller_env, concurrent_timeout, max_concurrency, **kwargs
            )
        )

    else:
        for gusty_dag in gusty_dags:
            dag_id = os.path.basename(gusty_dag)
            if current_dag_id is None or current_dag_id == dag_id:
                caller_env[dag_id] = create_dag(gusty_dag, **kwargs)



import yaml, nbformat, jupytext
from gusty.parsing.loaders import generate_loader
from gusty.importing import airflow_version
from absql.files.parsers import frontmatter_load
from gusty.parsing.utils import render_frontmatter, get_callable_from_file


def parse_generic(file_path, loader=None, runner=None, render_on_create=False):
    if loader is None:
        loader = generate_loader()
    # Read either the frontmatter or the parsed yaml file (using "or" to coalesce them)
    file_contents = frontmatter_load(file_path, loader=loader)
    job_spec = file_contents["metadata"] or file_contents["content"]

    if render_on_create:
        job_spec = render_frontmatter(job_spec, runner)

    return job_spec


def parse_py(file_path, loader=None, runner=None, render_on_create=False):
    if loader is None:
        loader = generate_loader()

    job_spec = {}
    if airflow_version > 1:
        job_spec.update({"operator": "airflow.operators.python.PythonOperator"})
    else:
        job_spec.update(
            {"operator": "airflow.operators.python_operator.PythonOperator"}
        )

    file_contents = jupytext.read(file_path)["cells"][0]

    # if spec contains metadata header...
    if file_contents["cell_type"] == "raw":
        assert (
            file_contents["source"] is not None
        ), "You need a comment block starting and ending with '# ---' at the top of {file_path}".format(
            file_path=file_path
        )
        assert (  # noqa
            "---" in file_contents["source"],
        ), "You need a comment block starting and ending with '# ---' at the top of {file_path}".format(
            file_path=file_path
        )
        source = file_contents["source"].replace("---", "")
        settings = yaml.load(source, Loader=loader)
        job_spec.update(**settings)

        # search for a python callable if one is specified
        if "python_callable" in job_spec.keys():
            callable_name = job_spec["python_callable"]
            job_spec.update(
                {"python_callable": get_callable_from_file(file_path, callable_name)}
            )
        # Default to sourcing this file for a PythonOperator
        else:
            job_spec.update({"python_callable": lambda: exec(open(file_path).read())})

        # Support additional callables
        if "extra_callables" in job_spec.keys():
            for arg_name, callable_name in job_spec["extra_callables"].items():
                job_spec.update(
                    {arg_name: get_callable_from_file(file_path, callable_name)}
                )

    # If no metadata then we also default to sourcing this file for a PythonOperator
    else:
        job_spec.update({"python_callable": lambda: exec(open(file_path).read())})

    if render_on_create:
        job_spec = render_frontmatter(job_spec, runner)

    return job_spec


def parse_ipynb(file_path, loader=None, runner=None, render_on_create=False):
    if loader is None:
        loader = generate_loader()
    # Find first yaml cell in jupyter notebook and parse yaml
    file_contents = nbformat.read(file_path, as_version=4)["cells"]
    yaml_cell = [
        cell
        for cell in file_contents
        if cell["cell_type"] == "markdown"
        and cell["source"].startswith(("```yaml", "```yml"))
    ]
    assert len(yaml_cell) > 0, "Please add a yaml block to %s" % file_path
    yaml_cell = yaml_cell[0]["source"]
    job_spec = yaml.load(
        yaml_cell.replace("```yaml", "").replace("```yml", "").replace("```", ""),
        Loader=loader,
    )

    if render_on_create:
        job_spec = render_frontmatter(job_spec, runner)

    return job_spec


def parse_sql(file_path, loader=None, runner=None, render_on_create=False):
    if loader is None:
        loader = generate_loader()
    file_contents = frontmatter_load(file_path, loader=loader)
    job_spec = file_contents["metadata"]
    job_spec["sql"] = file_contents["content"]

    if render_on_create:
        job_spec = render_frontmatter(job_spec, runner)

    return job_spec



# from airflow import macros
# from airflow.utils.context import VariableAccessor

from gusty.parsing.loaders import (
    default_constructors,
    handle_user_constructors,
)

# airflow_default_macros = {
#     # Cannot include macros because Airflow uses deepcopy
#     # "macros": macros,
#     "var": {
#         "json": VariableAccessor(deserialize_json=True),
#         "value": VariableAccessor(deserialize_json=False),
#     },
# }


def generate_loader_constructors(user_defined_macros, dag_constructors):
    loader_constructors = {}
    loader_constructors.update(default_constructors)
    loader_constructors.update(handle_user_constructors(dag_constructors))

    # user-defined macros are always preferred over
    # any default_constructors or dag_constructors,
    # and have to be formatted for YAML tags
    user_defined_constructors = {}
    for tag, func in user_defined_macros.items():
        tag = "!" + tag
        user_defined_constructors[tag] = func
    loader_constructors.update(user_defined_constructors)

    return loader_constructors


def generate_runner_context(loader_constructors):
    # Assumes that consolidation has already
    # occurred via generate_loader_constructors
    runner_context = {k.strip("!"): v for k, v in loader_constructors.items()}
    # support for airflow_default_macros, as long as the
    # user hasn't already added them via in user_defined_macros
    # or dag_constructors
    # for k, v in airflow_default_macros.items():
    #     if k not in runner_context.keys():
    #         runner_context[k] = v
    return runner_context


def generate_user_defined_macros(user_defined_macros, dag_constructors):
    macros_plus_constructors = {}
    macros_plus_constructors.update(
        {k.strip("!"): v for k, v in handle_user_constructors(dag_constructors).items()}
    )
    macros_plus_constructors.update(user_defined_macros)
    return macros_plus_constructors



import pytest
from datetime import datetime, timedelta
from gusty.utils import days_ago
from gusty import create_dag


def custom_days_ago():
    return days_ago(17)


def custom_retries(*args):
    retries = 0
    for a in args:
        retries += a
    return retries


@pytest.fixture(scope="session")
def loader_dir():
    return "tests/dags/loader"


@pytest.fixture(scope="session")
def dag(loader_dir):
    dag = create_dag(
        loader_dir,
        description="A dag created without metadata",
        schedule="0 0 * * *",
        default_args={
            "owner": "gusty",
            "depends_on_past": False,
            "start_date": days_ago(1),
            "email": "gusty@gusty.com",
            "email_on_failure": False,
            "email_on_retry": False,
            "retries": 3,
            "retry_delay": timedelta(minutes=5),
        },
        latest_only=False,
        dag_constructors=[custom_days_ago, custom_retries],
    )
    return dag


@pytest.fixture(scope="session")
def task(dag):
    return dag.__dict__["task_dict"]["loader_task"]


@pytest.fixture(scope="session")
def custom_task(dag):
    return dag.__dict__["task_dict"]["custom_task"]


@pytest.fixture(scope="session")
def sql_task(dag):
    sql_task = dag.task_dict["sql_task"]
    return sql_task


def test_datetime(task):
    start_date = task.__dict__["start_date"]
    assert start_date.date() == datetime(2022, 1, 15).date()


def test_days_ago(task):
    end_date = task.__dict__["end_date"]
    assert end_date == days_ago(4)


def test_timedelta(task):
    retry_delay = task.__dict__["retry_delay"]
    assert retry_delay == timedelta(minutes=20)


def test_retries(task):
    retries = task.__dict__["retries"]
    assert retries == 7


def test_custom_days_ago(custom_task):
    start_date = custom_task.__dict__["start_date"]
    assert start_date == days_ago(17)


def test_custom_retries(custom_task):
    retries = custom_task.__dict__["retries"]
    assert retries == 9


def test_sql_task_start(sql_task):
    start_date = sql_task.__dict__["start_date"]
    assert start_date == days_ago(24)


def test_sql_task_end(sql_task):
    end_date = sql_task.__dict__["end_date"]
    assert end_date == days_ago(17)



from gusty.building import GustyBuilder


def create_dag(
    dag_dir,
    task_group_defaults=None,
    wait_for_defaults=None,
    latest_only=True,
    parse_hooks=None,
    dag_constructors=None,
    extra_tags=None,
    render_on_create=False,
    **kwargs
):
    """
    Create an Airflow DAG by passing a directory of .yml task files.

    Each .yml task file should represent one task in the DAG. .ipynb and .Rmd files with
    yaml frontmatter are also accepted by create_dag. Any DAG kwargs (e.g. schedule_interval, description)
    can be passed directly into create_dag. Additionally, if your directory contains multiple subdirectories
    of .yml task files, and you are on Airflow V2, create_dag will create task groups for each subdirectory.
    By default, prefix_group_id is set to False.

    Parameters:
        dag_dir (str): A path to a directory of .yml task files.
            .ipynb and .Rmd files with yaml frontmatter are also accepted

        task_group_defaults (dict): Any parameter that could be passed to Airflow's TaskGroup class

        wait_for_defaults (dict): Any parameter that could be passed to Airflow's ExternalTaskSensor
            (which is what gusty users to create external_dependency tasks)

        latest_only (bool): When True, creates a latest_only operator at the root of the DAG, which will
            ensure tasks only run for the most recent run date

        parse_hooks (dict): Keys are file extensions (beginning with periods), values are functions that
            parse a given file path for a the associated file extension.

        render_on_create (bool): When set to True, gusty will render templated Jinja during dag
            creation. This is contrary to Airflow's default behavior, which doesn't render templates
            until runtime. This feature is disabled by default.

        kwargs: Any additional keyword argument that can be passed to Airflow's DAG class,
            e.g. default_args, schedule_interval, description

    Returns:
        dag (airflow.DAG): A fully initialized Airflow DAG, complete with tasks, task groups, dependencies,
            external dependencies, and latest_only operator (if wanted).
    """

    setup = GustyBuilder(
        dag_dir,
        task_group_defaults=task_group_defaults or {},
        wait_for_defaults=wait_for_defaults or {},
        latest_only=latest_only,
        parse_hooks=parse_hooks or {},
        dag_constructors=dag_constructors or {},
        extra_tags=extra_tags or [],
        render_on_create=render_on_create,
        **kwargs
    )
    [setup.parse_metadata(level) for level in setup.levels]
    [setup.check_metadata(level) for level in setup.levels]
    [setup.create_structure(level) for level in setup.levels]
    [setup.read_specs(level) for level in setup.levels]
    [setup.create_tasks(level) for level in setup.levels]
    [setup.create_level_dependencies(level) for level in setup.levels]
    [setup.create_task_dependencies(level) for level in setup.levels]
    [setup.create_task_external_dependencies(level) for level in setup.levels]
    [setup.create_level_external_dependencies(level) for level in setup.levels]
    [setup.create_leaf_tasks(level) for level in setup.levels]
    [setup.create_root_dependencies(level) for level in setup.levels]
    return setup.return_dag()


from gusty.multi import create_dags  # noqa



import ast, importlib.util
from absql.render import render_context


def render_frontmatter(frontmatter, runner=None, exclude=["sql"]):
    if runner is not None:
        renderable_frontmatter = {
            k: v for k, v in frontmatter.items() if k not in exclude
        }
        rendered_frontmatter = render_context(
            file_contents=renderable_frontmatter, extra_context=runner.extra_context
        )
        frontmatter.update(rendered_frontmatter)
    return frontmatter


def get_callable_from_file(file_path, callable_name):
    v = CallableFinder(callable_name)
    with open(file_path) as f:
        tree = ast.parse(f.read())
        v.visit(tree)
        if v.has_callable:
            mod_file = importlib.util.spec_from_file_location(
                "".join(i for i in file_path if i.isalnum()), file_path
            )
            mod = importlib.util.module_from_spec(mod_file)
            mod_file.loader.exec_module(mod)
            return getattr(mod, callable_name)
        else:
            assert (
                False
            ), "{file_path} specifies python_callable {callable} but {callable} not found in {file_path}".format(  # noqa
                file_path=file_path, callable=callable_name
            )


class CallableFinder(ast.NodeVisitor):
    def __init__(self, callable_name):
        self.has_callable = None
        self.callable_name = callable_name

    def visit_FunctionDef(self, node):
        ast.NodeVisitor.generic_visit(self, node)
        if node.name == self.callable_name:
            self.has_callable = True



import pytest
import os
from datetime import timedelta
from click.testing import CliRunner

from airflow import DAG
from airflow.operators.python import PythonOperator
from airflow.operators.bash import BashOperator
from airflow.providers.sqlite.operators.sqlite import SqliteOperator

from gusty.cli import cli, sample_tasks
from gusty import create_dag


@pytest.fixture()
def dag_name():
    return "cli_dag"


@pytest.fixture(scope="session")
def output_path(tmp_path_factory):
    return tmp_path_factory.mktemp("cli_output")


@pytest.fixture()
def cli_runner():
    return CliRunner()


@pytest.fixture()
def cli_result(dag_name, cli_runner, output_path):
    with cli_runner.isolated_filesystem(temp_dir=output_path):
        result = cli_runner.invoke(
            cli,
            ["use", "create-dag", f"--name={dag_name}", f"--dags-dir={output_path}"],
        )
    return result


@pytest.fixture()
def dag(output_path, dag_name):
    dag_path = os.path.join(output_path, dag_name)
    return create_dag(dag_path)


def test_cli_result(cli_result):
    assert cli_result.exit_code == 0


def test_create_dag_file(dag_name, output_path):
    # Get expected contents of the create_dag file
    (
        create_dag_filename,
        create_dag_file_expected_contents_list,
    ) = sample_tasks.create_dag_file(dag_name)
    create_dag_file_expected_contents = "".join(create_dag_file_expected_contents_list)

    # Check that the CLI-made create_dag file has expected contents
    create_dag_file_path = os.path.join(output_path, create_dag_filename)
    assert os.path.exists(create_dag_file_path)

    with open(create_dag_file_path, "r") as f:
        create_dag_file_contents = f.read()
        assert create_dag_file_contents == create_dag_file_expected_contents


def test_task_files(dag_name, output_path):
    dag_path = os.path.join(output_path, dag_name)
    assert os.path.exists(dag_path)

    for (
        task_file_name,
        task_file_expected_contents_list,
    ) in sample_tasks.dag_contents_map.items():
        # Get expected task file contents
        task_file_expected_contents = "".join(task_file_expected_contents_list)
        task_file_path = os.path.join(dag_path, task_file_name)
        assert os.path.exists(task_file_path)

        # Check that task files have expected contents
        with open(task_file_path, "r") as f:
            task_file_contents = f.read()
            assert task_file_contents == task_file_expected_contents


def test_dag_object(dag):
    assert isinstance(dag, DAG)


def test_python_task(dag):
    assert isinstance(dag.task_dict["hi"], PythonOperator)
    assert dag.task_dict["hi"].python_callable() == "hi"


def test_sql_task(dag):
    assert isinstance(dag.task_dict["hey"], SqliteOperator)
    assert dag.task_dict["hey"].sql == "SELECT 'hey'"


def test_yml_task(dag):
    assert isinstance(dag.task_dict["hello"], BashOperator)
    assert dag.task_dict["hello"].bash_command == "echo hello"


def test_latest_only_root(dag):
    assert len(dag.roots) == 1
    assert dag.roots[0].task_id == "latest_only"


def test_metadata(dag):
    assert dag.__dict__["default_args"]["retry_delay"] == timedelta(minutes=5)



import pytest
from gusty import create_dag
from gusty.utils import days_ago
from datetime import timedelta

###############
## FIXTURES ##
###############


@pytest.fixture(scope="session")
def multi_task_dir():
    return "tests/dags/multi_task"


@pytest.fixture(scope="session")
def dag(multi_task_dir):
    dag = create_dag(
        multi_task_dir,
        description="A dag created without metadata",
        schedule="0 0 * * *",
        default_args={
            "owner": "gusty",
            "depends_on_past": False,
            "start_date": days_ago(1),
            "email": "gusty@gusty.com",
            "email_on_failure": False,
            "email_on_retry": False,
            "retries": 3,
            "retry_delay": timedelta(minutes=5),
        },
        latest_only=False,
    )
    return dag


###########
## TESTS ##
###########


def test_multiple_tasks_exist(dag):
    assert "bash_a" in dag.task_dict.keys()
    assert "bash_b" in dag.task_dict.keys()
    assert "bash_c" in dag.task_dict.keys()
    assert "multi_bash" not in dag.task_dict.keys()


def test_multiple_tasks_changed(dag):
    bash_a_cmd = dag.task_dict["bash_a"].__dict__["bash_command"]
    bash_b_cmd = dag.task_dict["bash_b"].__dict__["bash_command"]
    assert bash_a_cmd == "echo a"
    assert bash_b_cmd == "echo b"


def test_multiple_tasks_defaults(dag):
    bash_c_cmd = dag.task_dict["bash_c"].__dict__["bash_command"]
    assert bash_c_cmd == "echo default"


def test_multiple_python_tasks_exist(dag):
    assert "python_a" in dag.task_dict.keys()
    assert "python_b" in dag.task_dict.keys()


def test_multiple_python_task_results(dag):
    callable_a = dag.task_dict["python_a"].__dict__["python_callable"]
    callable_b = dag.task_dict["python_b"].__dict__["python_callable"]
    assert callable_a() == "a"
    assert callable_b() == "b"


def test_combined_multi_partials(dag):
    python_email_a = dag.task_dict["python_a"].__dict__["email"]
    python_email_b = dag.task_dict["python_b"].__dict__["email"]
    assert python_email_a == "gusty@gusty.com"
    assert python_email_b == "multi@gusty.com"


def test_correct_task_count(dag):
    assert len(dag.task_dict) == 7


def test_nested_update(dag):
    op_kwargs = dag.task_dict["nested_py"].__dict__["op_kwargs"]
    assert op_kwargs["foo"] == "bar"
    assert op_kwargs["biz"] == "bazz"
    assert op_kwargs["min"] == "mouse"
    assert op_kwargs["outer_dict"] == {"outer": "dict"}
    assert op_kwargs["inner_dict"] == {"inner": "dict"}
    assert op_kwargs["updated_dict"] == {"updated": "updated"}


def test_nested_result(dag):
    op_kwargs = dag.task_dict["nested_py"].__dict__["op_kwargs"]
    callable = dag.task_dict["nested_py"].__dict__["python_callable"]
    assert callable(**op_kwargs) == "bar, bazz, mouse, dict, dict, updated"



import os, inspect
from airflow.models import BaseOperator, DAG
from absql import Runner
from functools import partial
from gusty.errors import NonexistentDagDirError
from gusty.parsing import parse, default_parsers
from gusty.parsing.loaders import generate_loader
from gusty.importing import airflow_version, get_operator
from gusty.external_sensor import create_external_sensor, make_external_task_name
from gusty.utils.context import (
    generate_loader_constructors,
    generate_runner_context,
    generate_user_defined_macros,
)

###########################
## Version Compatability ##
###########################

if airflow_version > 1:
    from airflow.utils.task_group import TaskGroup

if airflow_version > 1:
    from airflow.operators.latest_only import LatestOnlyOperator
else:
    from airflow.operators.latest_only_operator import LatestOnlyOperator

###############
## Constants ##
###############

AVAILABLE_WAIT_FOR_PARAMS = [
    "poke_interval",
    "timeout",
    "retries",
    "mode",
    "soft_fail",
    "execution_delta",
    "execution_date_fn",
    "check_existence",
]

BASE_OPERATOR_KEYS = inspect.signature(BaseOperator.__init__).parameters.keys()

#########################
## Schematic Functions ##
#########################


def create_schematic(dag_dir, parsers=default_parsers):
    """
    Given a dag directory, identify requirements (e.g spec_paths, metadata) for each "level" of the DAG.
    """
    return {
        # Each entry is a "level" of the main DAG
        os.path.abspath(dir): {
            "name": os.path.basename(dir),
            "parent_id": os.path.abspath(os.path.dirname(dir))
            if os.path.basename(os.path.dirname(dir))
            != os.path.basename(os.path.dirname(dag_dir))
            else None,
            "structure": None,
            "spec_paths": [
                os.path.abspath(os.path.join(dir, file))
                for file in files
                if file.endswith(tuple(parsers.keys()))
                and file != "METADATA.yml"
                and not file.startswith(("_", "."))
            ],
            "specs": [],
            "metadata_path": os.path.abspath(os.path.join(dir, "METADATA.yml"))
            if "METADATA.yml" in files
            else None,
            "metadata": {},
            "tasks": {},
            "dependencies": []
            if os.path.basename(os.path.dirname(dir))
            != os.path.basename(os.path.dirname(dag_dir))
            else None,
            "external_dependencies": [],
        }
        for dir, subdirs, files in os.walk(dag_dir)
        if not os.path.basename(dir).startswith(("_", "."))
    }


def get_level_structure(level_id, schematic):
    """
    Helper function to pull out a DAG level's structure (e.g. DAG or TaskGroup)
    """
    level_structure = schematic[level_id]["structure"]
    return level_structure


def get_top_level_dag(schematic):
    """
    Helper function to pull out the primary DAG object in a schematic
    """
    top_level_id = list(schematic.keys())[0]
    top_level_dag = get_level_structure(top_level_id, schematic)
    return top_level_dag


def parse_external_dependencies(external_dependencies):
    """
    Document
    """
    if isinstance(external_dependencies, dict):
        deps = []
        for dag_name, dag_details in external_dependencies.items():
            for task_name in dag_details["tasks"]:
                deps.append((dag_name, task_name))

    else:
        deps = [j for i in external_dependencies for j in i.items()]
    return deps


def parse_wait_for_overrides(external_dependencies):
    if isinstance(external_dependencies, dict):
        dags = external_dependencies.keys()
        dag_wait_for_overrides = {}
        for dag in dags:
            wait_for_specs = external_dependencies[dag].copy()
            wait_for_specs = {k: v for k, v in wait_for_specs.items() if k != "tasks"}
            dag_wait_for_overrides.update({dag: wait_for_specs})
        return dag_wait_for_overrides
    else:
        return {}


#######################
## Builder Functions ##
#######################


def _get_operator_parameters(operator, operator_param_cache):
    params = operator_param_cache.get(operator)
    if params:
        return params

    params = getattr(operator, "_gusty_parameters", None)
    if params is not None:
        operator_param_cache.update({operator: params})
        return params

    params = inspect.signature(operator.__init__).parameters.keys()
    operator_param_cache.update({operator: params})
    return params


def build_task(spec, level_id, schematic, operator_cache, operator_param_cache):
    """
    Given a task specification ("spec"), locate the operator and
    instantiate the object with args from the spec.
    """
    operator = get_operator(spec["operator"], operator_cache)

    args = {
        k: v
        for k, v in spec.items()
        if k in BASE_OPERATOR_KEYS
        or k in _get_operator_parameters(operator, operator_param_cache)
        or k in _get_operator_parameters(operator.__base__, operator_param_cache)
    }
    args["task_id"] = spec["task_id"]
    args["dag"] = get_top_level_dag(schematic)
    if airflow_version > 1:
        level_structure = get_level_structure(level_id, schematic)
        if isinstance(level_structure, TaskGroup):
            args["task_group"] = level_structure

    task = operator(**args)

    return task


def build_structure(schematic, parent_id, name, metadata):
    """
    Builds a DAG or a TaskGroup, contingent on the parent_id (which is None for the main DAG in a schematic)
    """
    is_top_level = parent_id is None
    if is_top_level:
        if metadata is not None:
            level_init_data = {
                k: v
                for k, v in metadata.items()
                if k in k in inspect.signature(DAG.__init__).parameters.keys()
            }
        structure = DAG(name, **level_init_data)

    else:
        # What is the main DAG?
        top_level_dag = get_top_level_dag(schematic)

        # What is the parent structure?
        parent = schematic[parent_id]["structure"]

        # Set some TaskGroup defaults
        level_defaults = {
            "group_id": name,
            "prefix_group_id": False,
            "dag": top_level_dag,
        }

        # If the parent structure is another TaskGroup, add it as parent_group kwarg
        if isinstance(parent, TaskGroup):
            level_defaults.update({"parent_group": parent})

        # Read in any metadata
        if metadata is not None:
            # scrub for TaskGroup inits only
            level_init_data = {
                k: v
                for k, v in metadata.items()
                if k in k in inspect.signature(TaskGroup.__init__).parameters.keys()
                and k not in ["dag", "parent_group"]
            }
            level_defaults.update(level_init_data)

        structure = TaskGroup(**level_defaults)

    return structure


def flatten_nested_lists(x):
    if isinstance(x, list):
        return [a for i in x for a in flatten_nested_lists(i)]
    else:
        return [x]


##################
## GustyBuilder ##
##################


class GustyBuilder:
    def __init__(self, dag_dir, **kwargs):
        """
        Because DAGs can be multiple "levels" now, the GustyBuilder class is here to first
        create a "schematic" of the DAG's levels and all of the specs associated with
        each level, at which point it moves through the schematic to build each level's
        "structure" (DAG or TaskGroup), tasks, dependencies, and external_dependencies, so on,
        until the DAG is complete.
        """

        if not os.path.isdir(dag_dir):
            raise NonexistentDagDirError(
                "DAG directory {dag_dir} does not exist".format(dag_dir=dag_dir)
            )

        self.extra_tags = kwargs.get("extra_tags")

        self.render_on_create = kwargs.get("render_on_create")

        self.parsers = default_parsers.copy()

        if len(kwargs["parse_hooks"]) > 0:
            assert isinstance(
                kwargs["parse_hooks"], dict
            ), "parse_hooks should be a dict of file extensions and handler functions for file_path."
            self.parsers.update(kwargs["parse_hooks"])

        # ------------------------- #
        # Begin Consolidate Context #
        # ------------------------- #

        user_defined_macros = kwargs.get("user_defined_macros", {})
        user_dag_constructors = kwargs.get("dag_constructors", {})

        # Create dag_constructors (for the loader)
        dag_constructors = generate_loader_constructors(
            user_defined_macros=user_defined_macros,
            dag_constructors=user_dag_constructors,
        )

        # Create runner_context (for the runner)
        runner_context = generate_runner_context(dag_constructors)

        # Make the runner
        self.runner = Runner(**runner_context)

        # Make the loader
        self.loader = generate_loader(dag_constructors)

        # Consolidate user_defined_macros
        kwargs.update(
            {
                "user_defined_macros": generate_user_defined_macros(
                    user_defined_macros=user_defined_macros,
                    dag_constructors=user_dag_constructors,
                )
            }
        )

        # ----------------------- #
        # END Consolidate Context #
        # ----------------------- #

        self.schematic = create_schematic(dag_dir, self.parsers)

        # DAG defaults - everything that's not task_group_defaults or wait_for_defaults
        # is considered DAG default metadata
        self.dag_defaults = {
            k: v
            for k, v in kwargs.items()
            if k not in ["task_group_defaults", "wait_for_defaults"]
        }

        # TaskGroup defaults
        self.task_group_defaults = (
            kwargs["task_group_defaults"]
            if "task_group_defaults" in kwargs.keys()
            else {}
        )

        # external dependency / wait_for_defaults
        self.wait_for_defaults = {"poke_interval": 10, "timeout": 3600, "retries": 60}
        if "wait_for_defaults" in kwargs.keys():
            user_wait_for_defaults = {
                k: v
                for k, v in kwargs["wait_for_defaults"].items()
                if k in AVAILABLE_WAIT_FOR_PARAMS or k in BASE_OPERATOR_KEYS
            }
            self.wait_for_defaults.update(user_wait_for_defaults)

        if "wait_for_class" in kwargs.keys():
            self.create_external_sensor = partial(
                create_external_sensor, sensor=kwargs["wait_for_class"]
            )
        else:
            self.create_external_sensor = create_external_sensor

        # We will accept multiple levels only for Airflow v2 and up
        # This will keep the TaskGroup logic of the Levels class
        # Solely for Airflow v2 and beyond
        self.levels = [level_id for level_id in self.schematic.keys()]
        self.levels = [self.levels[0]] if airflow_version < 2 else self.levels

        # For tasks gusty creates outside of specs provided by the directory
        # It is important for gusty to keep a record  of the tasks created.
        # We keep a running list of all_tasks, as well.
        self.wait_for_tasks = {}
        self.all_tasks = {}

        # Allow for the assignment of additional leaf tasks from a dictionary
        self.leaf_tasks_from_dict = kwargs.get("leaf_tasks_from_dict", {})

        # caching
        self.operator_param_cache = {}
        self.operator_cache = {}

    def parse_metadata(self, id):
        """
        For a given level id, parse any metadata if there is a METADATA.yml path,
        otherwise add applicable metadata defaults for that level.
        """

        if self.schematic[id]["parent_id"] is None:
            metadata_defaults = self.dag_defaults.copy()
        else:
            metadata_defaults = self.task_group_defaults.copy()

        # METADATA.yml will override defaults
        level_metadata_path = self.schematic[id]["metadata_path"]
        if os.path.exists(level_metadata_path or ""):
            level_metadata = default_parsers[".yml"](
                level_metadata_path, self.loader, self.runner, render_on_create=True
            )

            # special case - default_args provided in both metadata_defaults and level_metadata
            if (
                self.schematic[id]["parent_id"] is None
                and "default_args" in metadata_defaults.keys()
                and "default_args" in level_metadata.keys()
            ):
                # metadata defaults
                metadata_default_args = metadata_defaults["default_args"].copy()
                # level_metadata (provided via METADATA.yml)
                level_default_args = level_metadata["default_args"].copy()
                # metadata defaults updated with level_metadata
                metadata_default_args.update(level_default_args)
                # updated and resolved attached back to level_metadata
                level_metadata.update({"default_args": metadata_default_args})

            # special case - wait_for_defaults
            if (
                self.schematic[id]["parent_id"] is None
                and "wait_for_defaults" in level_metadata.keys()
            ):
                self.wait_for_defaults.update(level_metadata["wait_for_defaults"])

        else:
            level_metadata = {}
        metadata_defaults.update(level_metadata)

        if self.extra_tags and self.schematic[id]["parent_id"] is None:
            tags = metadata_defaults.get("tags", [])
            tags = list(set(tags + self.extra_tags))
            metadata_defaults.update({"tags": tags})

        self.schematic[id]["metadata"] = metadata_defaults

        # dependencies get explicity set at the level-"level" for each level
        # and must be pulled out separately from other metadata.
        # metadata_default_dependencies allows for root-level default external dependencies
        # to be set at the root DAG level in create_dag. Any dependencies set in METADATA.yml
        # will override any defaults set in metadata_default_dependencies
        level_dependencies = {
            k: v
            for k, v in level_metadata.items()
            if k in k in ["dependencies", "external_dependencies"]
        }
        metadata_default_dependencies = {
            k: v
            for k, v in metadata_defaults.items()
            if k in k in ["external_dependencies"]
        }
        if len(level_dependencies) > 0:
            self.schematic[id].update(level_dependencies)
        # metadata_default_dependencies is really meant for the root level only
        elif (
            len(metadata_default_dependencies) > 0
            and self.schematic[id]["parent_id"] is None
            and "external_dependencies" in metadata_default_dependencies.keys()
        ):
            root_externals = metadata_default_dependencies["external_dependencies"]
            assert isinstance(
                root_externals, list
            ), """Root external dependencies set in create_dag must be a list of dicts following the pattern {"dag_id": "task_id"}"""  # noqa
            assert all(
                [isinstance(dep, dict) for dep in root_externals]
            ), """Root external dependencies set in create_dag must be a list of dicts following the pattern {"dag_id": "task_id"}"""  # noqa
            self.schematic[id].update({"external_dependencies": root_externals})

    def check_metadata(self, id):
        if id in self.levels:
            # ignore subfolders
            if self.schematic[id]["parent_id"] is None:
                ignore_subfolders = self.schematic[id]["metadata"].get(
                    "ignore_subfolders", False
                )
                if ignore_subfolders:
                    self.schematic = {
                        k: v
                        for k, v in self.schematic.items()
                        if self.schematic[k]["parent_id"] is None
                    }
                    self.levels = [level_id for level_id in self.schematic.keys()]

    def create_structure(self, id):
        """
        Given a level of the DAG, a structure such as a DAG or a TaskGroup will be initialized.
        """
        level_schematic = self.schematic[id]
        level_kwargs = {
            k: v
            for k, v in level_schematic.items()
            if k in inspect.signature(build_structure).parameters.keys()
        }

        level_structure = build_structure(self.schematic, **level_kwargs)

        # We update the schematic with the structure of the level created
        self.schematic[id].update({"structure": level_structure})

    def read_specs(self, id):
        """
        For a given level id, parse all of that level's yaml specs, given paths to those files.
        """
        level_metadata = self.schematic[id]["metadata"]
        level_spec_paths = self.schematic[id]["spec_paths"]
        level_specs = [
            parse(
                spec_path, self.parsers, self.loader, self.runner, self.render_on_create
            )
            for spec_path in level_spec_paths
        ]
        level_specs = flatten_nested_lists(level_specs)

        # Add any additional specs from leaf_tasks_from_dict
        if self.schematic[id]["parent_id"] is None and self.leaf_tasks_from_dict:
            # Existing specs will take precedence over any from the dict
            existing_task_ids = [s["task_id"] for s in level_specs]
            for task_id, task_spec in self.leaf_tasks_from_dict.items():
                if task_id.lower().strip() not in existing_task_ids:
                    task_spec.update({"task_id": task_id.lower().strip()})
                    level_specs.append(task_spec)

        if airflow_version > 1:
            level_structure = self.schematic[id]["structure"]
            level_name = self.schematic[id]["name"]
            add_suffix = (
                level_metadata["suffix_group_id"]
                if "suffix_group_id" in level_metadata.keys()
                else False
            )
            if isinstance(level_structure, TaskGroup):
                if level_structure.prefix_group_id and not add_suffix:
                    for level_spec in level_specs:
                        level_spec["task_id"] = "{x}_{y}".format(
                            x=level_name, y=level_spec["task_id"]
                        )
                elif add_suffix:
                    for level_spec in level_specs:
                        level_spec["task_id"] = "{y}_{x}".format(
                            x=level_name, y=level_spec["task_id"]
                        )
        self.schematic[id].update({"specs": level_specs})

    def create_tasks(self, id):
        """
        For a given level id, create all tasks based on the specs parsed from read_specs
        """
        level_specs = self.schematic[id]["specs"]
        level_tasks = {
            spec["task_id"]: build_task(
                spec, id, self.schematic, self.operator_cache, self.operator_param_cache
            )
            for spec in level_specs
        }
        self.schematic[id]["tasks"] = level_tasks
        self.all_tasks.update(level_tasks)

    def create_level_dependencies(self, id):
        """
        For a given level id, identify what would be considered a valid set of dependencies within the DAG
        for that level, and then set any specified dependencies upstream of that level. An example here would
        be a DAG has two .yml jobs and one subfolder, which (the subfolder) is turned into a TaskGroup. That
        TaskGroup can validly depend on the two .yml jobs, so long as either of those task_ids are defined
        within the dependencies section of the TaskGroup's METADATA.yml
        """
        level_structure = self.schematic[id]["structure"]
        level_dependencies = self.schematic[id]["dependencies"]
        level_parent_id = self.schematic[id]["parent_id"]
        if level_parent_id is not None:
            level_parent = self.schematic[level_parent_id]
            parent_tasks = level_parent[
                "tasks"
            ]  # these follow the format {task_id: task_object}
            sibling_levels = {
                level["name"]: level["structure"]
                for level_id, level in self.schematic.items()
                if level["parent_id"] == level_parent_id and level_id != id
            }  # these follow the format {level_name: structure_object}
            valid_dependency_objects = {**parent_tasks, **sibling_levels}
            valid_dependency_objects = {
                k: v for k, v in valid_dependency_objects.items() if v is not None
            }
            for dependency in level_dependencies:
                if dependency in valid_dependency_objects.keys():
                    level_structure.set_upstream(valid_dependency_objects[dependency])

    def create_task_dependencies(self, id):
        """
        For a given level id, identify what would be considered a valid set of dependencies within the dag
        for that level, and then set any specified dependencies for each task at that level, as specified by
        the task's specs.
        """
        level_specs = self.schematic[id]["specs"]
        level_tasks = self.schematic[id]["tasks"]
        level_structure = self.schematic[id]["structure"]
        level_metadata = self.schematic[id]["metadata"]
        sibling_levels = {
            level["name"]: level["structure"]
            for level_id, level in self.schematic.items()
            if level["parent_id"] == id and level_id != id
        }
        valid_dependency_objects = {**self.all_tasks, **sibling_levels}
        valid_dependency_objects = {
            k: v for k, v in valid_dependency_objects.items() if v is not None
        }
        for task_id, task in level_tasks.items():
            # task_dependencies allows users to add a dependencies attribute to
            # their custom operators, so that dependencies could be automatically detected
            # through something like regex detecting schemas in a SQL query
            task_dependencies = (
                task.dependencies if hasattr(task, "dependencies") else []
            )
            spec_dependencies = {
                task_id: spec["dependencies"]
                for spec in level_specs
                if spec["task_id"] == task_id and "dependencies" in spec.keys()
            }
            if len(spec_dependencies) > 0:
                spec_dependencies = spec_dependencies[task_id]

                # special case: prefixing or suffixing in task groups
                # may create a case where a spec dependency is incorrectly named;
                # the intuitive thing to happen is that gusty can pick up name changes
                # for items in the same task group
                if airflow_version > 1:
                    if isinstance(level_structure, TaskGroup):
                        add_suffix = (
                            level_metadata["suffix_group_id"]
                            if "suffix_group_id" in level_metadata.keys()
                            else False
                        )
                        add_prefix = level_structure.prefix_group_id
                        if add_prefix or add_suffix:
                            potential_task_names = (
                                [
                                    "{x}_{y}".format(
                                        x=self.schematic[id]["name"], y=dep
                                    )
                                    for dep in spec_dependencies
                                ]
                                if add_prefix
                                else [
                                    "{x}_{y}".format(
                                        x=dep, y=self.schematic[id]["name"]
                                    )
                                    for dep in spec_dependencies
                                ]
                            )

                            for potential_name in potential_task_names:
                                if potential_name in valid_dependency_objects.keys():
                                    spec_dependencies.append(potential_name)

                spec_task_dependencies = task_dependencies + spec_dependencies
            else:
                spec_task_dependencies = task_dependencies

            spec_task_dependencies = list(set(spec_task_dependencies))

            spec_task_dependencies = [
                dependency
                for dependency in spec_task_dependencies
                if dependency in valid_dependency_objects.keys()
            ]

            if len(spec_task_dependencies) > 0:
                for dependency in spec_task_dependencies:
                    task.set_upstream(valid_dependency_objects[dependency])

    def create_task_external_dependencies(self, id):
        """
        As with create_task_dependencies, for a given level id, parse all of the
        external dependencies in a task's spec, then create and add those "wait_for_"
        tasks upstream of a given task. Note the Builder class must keep a record of all
        "wait_for_" tasks as to not recreate the same task twice.
        """
        level_specs = self.schematic[id]["specs"]
        level_tasks = self.schematic[id]["tasks"]
        for task_id, task in level_tasks.items():
            task_spec_external_dependencies = {
                task_id: spec["external_dependencies"]
                for spec in level_specs
                if spec["task_id"] == task_id and "external_dependencies" in spec.keys()
            }
            if len(task_spec_external_dependencies) > 0:
                task_external_dependencies = parse_external_dependencies(
                    task_spec_external_dependencies[task_id]
                )
                wait_for_overrides = parse_wait_for_overrides(
                    task_spec_external_dependencies[task_id]
                )
                for dag_task_pair in task_external_dependencies:
                    (external_dag_id, external_task_id) = dag_task_pair

                    wait_for_task_name = make_external_task_name(
                        external_dag_id, external_task_id
                    )

                    if wait_for_task_name in self.wait_for_tasks.keys():
                        wait_for_task = self.wait_for_tasks[wait_for_task_name]
                        task.set_upstream(wait_for_task)
                    else:
                        wait_for_task = self.create_external_sensor(
                            get_top_level_dag(self.schematic),
                            external_dag_id,
                            external_task_id,
                            self.wait_for_defaults,
                            wait_for_overrides,
                            wait_for_task_name,
                        )
                        self.wait_for_tasks.update({wait_for_task_name: wait_for_task})
                        task.set_upstream(wait_for_task)

    def create_level_external_dependencies(self, id):
        """
        Same as create_task_external_dependencies, except for levels intead of tasks
        """
        level_structure = self.schematic[id]["structure"]
        level_external_dependencies = self.schematic[id]["external_dependencies"]
        level_parent_id = self.schematic[id]["parent_id"]
        if level_parent_id is not None:
            if len(level_external_dependencies) > 0:
                level_external_dependencies_parsed = parse_external_dependencies(
                    level_external_dependencies
                )
                wait_for_overrides = parse_wait_for_overrides(
                    level_external_dependencies
                )
                for dag_task_pair in level_external_dependencies_parsed:
                    (external_dag_id, external_task_id) = dag_task_pair

                    wait_for_task_name = make_external_task_name(
                        external_dag_id, external_task_id
                    )

                    if wait_for_task_name in self.wait_for_tasks.keys():
                        wait_for_task = self.wait_for_tasks[wait_for_task_name]
                        level_structure.set_upstream(wait_for_task)
                    else:
                        wait_for_task = self.create_external_sensor(
                            get_top_level_dag(self.schematic),
                            external_dag_id,
                            external_task_id,
                            self.wait_for_defaults,
                            wait_for_overrides,
                            wait_for_task_name,
                        )
                        self.wait_for_tasks.update({wait_for_task_name: wait_for_task})
                        level_structure.set_upstream(wait_for_task)

    def create_root_dependencies(self, id):
        """
        Finally, we look at the root level for a latest only spec. If latest_only, we create
        latest only and wire up any tasks and groups from the parent level to these deps if needed.
        """
        level_metadata = self.schematic[id]["metadata"]
        level_root_tasks = (
            level_metadata["root_tasks"]
            if "root_tasks" in level_metadata.keys()
            else None
        )
        level_parent_id = self.schematic[id]["parent_id"]
        level_external_dependencies = self.schematic[id]["external_dependencies"]
        # parent level only
        if level_parent_id is None:
            # What are valid downstream tasks for root-level dependencies
            level_tasks = self.schematic[id]["tasks"]
            # Special case - consider task groups masquerading as tasks (e.g. composite task groups)
            if airflow_version > 1:
                level_tasks_tasks = {
                    k: v for k, v in level_tasks.items() if not isinstance(v, TaskGroup)
                }
                level_tasks_groups = {
                    k: v for k, v in level_tasks.items() if isinstance(v, TaskGroup)
                }
                # Remove task groups that depend on other task groups from valid_dependency_objects
                level_tasks_groups = {
                    k: v
                    for k, v in level_tasks_groups.items()
                    if len(v.upstream_group_ids) == 0
                }
                level_tasks = {**level_tasks_tasks, **level_tasks_groups}
            if airflow_version > 1:
                child_levels = {
                    level["name"]: level["structure"]
                    for level_id, level in self.schematic.items()
                    if level["parent_id"] == id
                    and len(level["structure"].upstream_group_ids) == 0
                }
            else:
                child_levels = {
                    level["name"]: level["structure"]
                    for level_id, level in self.schematic.items()
                    if level["parent_id"] == id
                }
            valid_dependency_objects = {
                **level_tasks,
                **child_levels,
                **self.wait_for_tasks,
            }

            valid_dependency_objects = {
                k: v for k, v in valid_dependency_objects.items() if v is not None
            }

            # Set any root-level tasks
            if level_root_tasks is not None:
                valid_root_tasks = {**level_tasks}
                valid_root_tasks = {
                    id: task
                    for id, task in valid_root_tasks.items()
                    if id in level_root_tasks
                }

                for task_id, task in valid_root_tasks.items():
                    assert (
                        len(task.upstream_task_ids) == 0
                    ), "Task {id} in DAG {dag} is listed as a root task, but has dependencies listed: {dep_list}".format(  # noqa
                        id=task_id,
                        dag=self.schematic[id]["structure"]._dag_id,
                        dep_list=", ".join(task.upstream_task_ids),
                    )
                    assert (
                        len(task.downstream_task_ids) == 0
                    ), "Task {id} in DAG {dag} is listed as a root task, but other tasks explicity depend on it: {dep_list}".format(  # noqa
                        id=task_id,
                        dag=self.schematic[id]["structure"]._dag_id,
                        dep_list=", ".join(task.downstream_task_ids),
                    )

                for name, dependency in valid_dependency_objects.items():
                    if len(dependency.upstream_task_ids) == 0 or all(
                        [
                            dep_id.startswith("wait_for_")
                            for dep_id in dependency.upstream_task_ids
                        ]
                    ):
                        for root_name, root_dep in valid_root_tasks.items():
                            if (
                                name != root_name
                                and name not in valid_root_tasks.keys()
                            ):
                                dependency.set_upstream(root_dep)

            # Set root-level external dependencies
            if len(level_external_dependencies) > 0:
                root_external_dependencies = {}
                level_external_dependencies_parsed = parse_external_dependencies(
                    level_external_dependencies
                )
                wait_for_overrides = parse_wait_for_overrides(
                    level_external_dependencies
                )
                for dag_task_pair in level_external_dependencies_parsed:
                    (external_dag_id, external_task_id) = dag_task_pair

                    wait_for_task_name = make_external_task_name(
                        external_dag_id, external_task_id
                    )

                    if wait_for_task_name in self.wait_for_tasks.keys():
                        wait_for_task = self.wait_for_tasks[wait_for_task_name]
                        root_external_dependencies.update(
                            {wait_for_task_name: wait_for_task}
                        )
                    else:
                        wait_for_task = self.create_external_sensor(
                            get_top_level_dag(self.schematic),
                            external_dag_id,
                            external_task_id,
                            self.wait_for_defaults,
                            wait_for_overrides,
                            wait_for_task_name,
                        )

                        self.wait_for_tasks.update({wait_for_task_name: wait_for_task})
                        root_external_dependencies.update(
                            {wait_for_task_name: wait_for_task}
                        )
                for name, dependency in valid_dependency_objects.items():
                    if len(dependency.upstream_task_ids) == 0:
                        for ext_name, ext_dep in root_external_dependencies.items():
                            if name != ext_name:
                                dependency.set_upstream(ext_dep)
                valid_dependency_objects.update(**root_external_dependencies)

            # Set latest only if latest only
            if level_metadata is not None:
                level_latest_only = (
                    level_metadata["latest_only"]
                    if "latest_only" in level_metadata.keys()
                    else True
                )
            if level_latest_only:
                latest_only_operator = LatestOnlyOperator(
                    task_id="latest_only", dag=get_top_level_dag(self.schematic)
                )
                for name, dependency in valid_dependency_objects.items():
                    if len(dependency.upstream_task_ids) == 0:
                        dependency.set_upstream(latest_only_operator)

    def create_leaf_tasks(self, id):
        level_parent_id = self.schematic[id]["parent_id"]
        if level_parent_id is None:
            level_metadata = self.schematic[id]["metadata"]
            level_leaf_tasks = (
                level_metadata["leaf_tasks"]
                if "leaf_tasks" in level_metadata.keys()
                else []
            )
            # Add leaf tasks from dict if applicable
            if self.leaf_tasks_from_dict:
                level_leaf_tasks.extend(list(self.leaf_tasks_from_dict.keys()))
                # Ensure unique values
                level_leaf_tasks = list(set(level_leaf_tasks))

            level_root_tasks = (
                level_metadata["root_tasks"]
                if "root_tasks" in level_metadata.keys()
                else []
            )
            if level_leaf_tasks:
                level_tasks = self.schematic[id]["tasks"]
                child_levels = {
                    level["name"]: level["structure"]
                    for level_id, level in self.schematic.items()
                    if level["parent_id"] == id
                }
                valid_dependency_objects = {
                    **level_tasks,
                    **child_levels,
                }
                valid_dependency_objects = {
                    k: v for k, v in valid_dependency_objects.items() if v is not None
                }

                valid_leaf_tasks = {**level_tasks}
                valid_leaf_tasks = {
                    id: task
                    for id, task in valid_leaf_tasks.items()
                    if id in level_leaf_tasks
                }

                for task_id, task in valid_leaf_tasks.items():
                    assert (
                        len(task.upstream_task_ids) == 0
                    ), "Task {id} in DAG {dag} is listed as a leaf task, but has dependencies listed: {dep_list}".format(  # noqa
                        id=task_id,
                        dag=self.schematic[id]["structure"]._dag_id,
                        dep_list=", ".join(task.upstream_task_ids),
                    )
                    assert (
                        len(task.downstream_task_ids) == 0
                    ), "Task {id} in DAG {dag} is listed as a leaf task, but other tasks explicity depend on it: {dep_list}".format(  # noqa
                        id=task_id,
                        dag=self.schematic[id]["structure"]._dag_id,
                        dep_list=", ".join(task.downstream_task_ids),
                    )

                for name, dependency in valid_dependency_objects.items():
                    if len(dependency.downstream_task_ids) == 0:
                        for leaf_name, leaf_dep in valid_leaf_tasks.items():
                            if (
                                name != leaf_name
                                and name not in valid_leaf_tasks.keys()
                                and name not in level_root_tasks
                            ):
                                leaf_dep.set_upstream(dependency)

    def return_dag(self):
        return get_top_level_dag(self.schematic)



import pytest
from gusty.importing import (
    get_operator_location,
    get_operator_name,
    get_operator_module,
    get_operator,
)


@pytest.fixture
def operator_string():
    return "airflow.operators.bash.BashOperator"


def test_location(operator_string):
    location = get_operator_location(operator_string)
    assert location == "airflow"


def test_operator_name(operator_string):
    operator_name = get_operator_name(operator_string)
    assert operator_name == "BashOperator"


def test_operator_module(operator_string):
    operator_module = get_operator_module(operator_string)
    assert operator_module == "airflow.operators.bash"


def test_module_fail():
    with pytest.raises(AssertionError):
        get_operator_module("BashOperator")


def test_get_operator():
    operator = get_operator("airflow.operators.bash.BashOperator", {})
    assert operator.__name__ == "BashOperator"



from gusty.importing import airflow_version

if airflow_version > 1:
    from airflow.sensors.external_task import ExternalTaskSensor
else:
    from airflow.sensors.external_task_sensor import ExternalTaskSensor


def create_external_sensor(
    dag,
    external_dag_id,
    external_task_id,
    wait_for_defaults,
    wait_for_overrides,
    wait_for_task_name,
    sensor=ExternalTaskSensor,
):
    if sensor.__name__ in ["EmptyOperator", "DummyOperator"]:
        return sensor(dag=dag, task_id=wait_for_task_name)

    ext_task_defaults = wait_for_defaults.copy()
    ext_task_overrides = wait_for_overrides.get(external_dag_id, {})
    # Only one of execution_delta or
    # execution_date_fn can be passed to
    # the ExternalTaskSensor
    if "execution_delta" in ext_task_overrides:
        if "execution_date_fn" in ext_task_defaults:
            ext_task_defaults.pop("execution_date_fn")
    ext_task_defaults.update(
        {
            **ext_task_overrides,
            "external_dag_id": external_dag_id,
            "external_task_id": (
                external_task_id if external_task_id != "all" else None
            ),
        }
    )

    wait_for_task = sensor(dag=dag, task_id=wait_for_task_name, **ext_task_defaults)

    return wait_for_task


def make_external_task_name(external_dag_id, external_task_id):
    return (
        "wait_for_DAG_{x}".format(x=external_dag_id)
        if external_task_id == "all"
        else "wait_for_{dag}_{task}".format(dag=external_dag_id, task=external_task_id)
    )



import pytest
from gusty import create_dag

##############
## FIXTURES ##
##############


@pytest.fixture(scope="session")
def leaves_from_dag_dir():
    return "tests/dags/leaves_from"


@pytest.fixture(scope="session")
def dag(leaves_from_dag_dir):
    dag = create_dag(
        leaves_from_dag_dir,
        leaf_tasks_from_dict={
            "dict_leaf": {"operator": "airflow.operators.empty.EmptyOperator"},
            "overwritten_leaf": {"operator": "airflow.operators.empty.EmptyOperator"},
        },
    )
    return dag


###########
## Tests ##
###########


def test_only_leaves_are_leaves(dag):
    task_dict = dag.__dict__["task_dict"]
    # These should not be leaves
    assert task_dict["python_a"] not in dag.leaves
    assert task_dict["python_b"] not in dag.leaves
    # These should be leaves
    assert task_dict["yaml_leaf"] in dag.leaves
    assert task_dict["dict_leaf"] in dag.leaves
    assert task_dict["overwritten_leaf"] in dag.leaves


def test_yaml_spec_overrides_leaf_dict(dag):
    assert (
        dag.__dict__["task_dict"]["overwritten_leaf"].__dict__["bash_command"]
        == "echo overwritten"
    )



import pytest
from datetime import timedelta
from gusty.utils import days_ago
from gusty import create_dag
from gusty.parsing import parse

##############
## FIXTURES ##
##############


@pytest.fixture(scope="session")
def parsing_dag_dir():
    return "tests/dags/parsing"


@pytest.fixture(scope="session")
def dag(parsing_dag_dir):
    dag = create_dag(
        parsing_dag_dir,
        description="A dag with some custom parsing functions.",
        schedule="0 0 * * *",
        default_args={
            "owner": "gusty",
            "depends_on_past": False,
            "start_date": days_ago(1),
            "email": "gusty@gusty.com",
            "email_on_failure": False,
            "email_on_retry": False,
            "retries": 3,
            "retry_delay": timedelta(minutes=5),
        },
        ignore_subfolders=True,
        parse_hooks={
            ".py": lambda file_path: {
                "operator": "airflow.operators.python.PythonOperator",
                "python_callable": lambda: "this was custom",
            },
        },
    )
    return dag


@pytest.fixture(scope="session")
def custom_task(dag):
    custom_task = dag.task_dict["a_parse_hook_task"]
    return custom_task


@pytest.fixture(scope="session")
def sql_task(dag):
    custom_task = dag.task_dict["sql_task"]
    return custom_task


###########
## Tests ##
###########


def test_read_yaml_spec():
    yaml_spec = parse("tests/dags/no_metadata/top_level_task.yml")
    assert yaml_spec["task_id"] == "top_level_task"
    assert yaml_spec["file_path"] == "tests/dags/no_metadata/top_level_task.yml"
    assert "operator" in yaml_spec.keys()
    assert "bash_command" in yaml_spec.keys()


def test_parse_hooks(custom_task):
    callable = custom_task.__dict__["python_callable"]
    res = callable()
    assert res == "this was custom"


def test_sql_parse(sql_task):
    assert sql_task.sql == "SELECT *\nFROM gusty_table"
